/* Output from p2c, the Pascal-to-C translator */
/* From input file "nvc.mod" */


#include <p2c/p2c.h>


/*
 *
 *      Noise vs. Chaos code.
 *
 *   This code implements a test to compare a given signal
 *   to the hypothesis that the signal is colored noise.  The test
 *   is designed to detect predictability in low dimensional dynamical
 *   systems, so if the test fails, the data may be from a
 *   chaotic system.  (Though not always! The test can be failed by
 *   nonlinear noise, I think, and periodic signals plus added noise)

 *   The method is in:
 *   M. B. Kennel and S. Isabelle, ``Method to distinguish possible
 *   chaos from colored noise and determine embedding parameters'',
 *   to be published in September 15th,1992 issue of Physical Review A.
 *   Preprint also available by anonymous FTP from 'lyapunov.ucsd.edu'.

 *      Usage:
 *
 *      nvc -N Npoints(Power of 2)
 *          -delay time_delay_min time_delay_max time_delay_step
 *          -pred prediction_time
 *          -correl decorrelation_dime
 *          -decimate prediction_errors_decimation_interval
 *          -nfake number_of_fake_datasets
 *          -dim embedding_dimension_min embedding_dimension_max

        (all those parameters are integers)

 *          -o output_file_name
 *
 *          -A datafile.ascii | -B datafile.dbin
 *      (input files, either in ASCII or double precision binary FP)
 *
 *      Optional:
 *           -phasesonly         Only randomize phases to make fake data,
 *         not amplitude.
             -dumpdata           Dump the gaussianized 1-d real and fake
                                 datasets to DUMP.dat--EVERYTIME
             -dumperrs Dump the predictor errors to DUMPERRS.dat
 */

#ifndef DynArray_H
#include "DynArray.h"
#endif

#ifndef DynMatrix_H
#include "DynMatrix.h"
#endif

#ifndef Util_H
#include "Util.h"
#endif

#ifndef KdTree_H
#include "KdTree.h"
#endif

#ifndef Math_H
#include "Math.h"
#endif


#ifndef IO_H
#include "IO.h"
#endif



Static float *OneDData;   /* The 1-D data array */

Static unsigned long Ndata;   /* the number of data points */

Static long iseed;


Static unsigned long Tau1, Tau2, DTau, PredTime, Dmin, Dmax, Correl, Decimate,
		     Nfake;

Static FILE *outfile;
Static Char outfilename[256];

Static boolean phasesonly;   /* do we want to randomize phases only */
Static boolean dumpdata;   /* Dump data sets */
Static boolean dumperrs;   /* Dump errors */


Static unsigned long nvc_NearestPowerOfTwo(unsigned long N)
{
    unsigned long i, j, k;

    j = N;
    i = 0;
    do {
	j /= 2.0;
	i++;
    } while (j != 1);

    for (k = 1; k <= i; k++)
	j *= 2;

    return j;

}


Static void nvc_ReadInData(void)
{
    /*
     *    From the arguments passed to the program get the input data vector
     *    Either from an ascii file, or from a binary file:
     *
     *    fnn -A filename.ascii
     *   or
     *    fnn -B filename.binary
     *
     *    -B is searched for FIRST on the command line
     *    Allocates OneDData;
     */
    double *doublearray;
    unsigned long argplace;
    FILE *binfile;
    FILE *asciifile;
    unsigned long i;
    double number;
    Char TEMP[256];
    Char STR1[256];
    unsigned long FORLIM;

    asciifile = NULL;
    binfile = NULL;
    argplace = Util_ArgvStringExists("-N");
    if (argplace == 0)
	Util_FatalError("You must provide a -N npoints flag!");

    strcpy(TEMP, P_argv[argplace + 1]);
    Ndata = Util_ScanInteger(TEMP);

    printf("Ndata = %12lu NearestPowerOfTwo(Ndata) = %12lu\n",
	   Ndata, nvc_NearestPowerOfTwo(Ndata));
    Ndata = nvc_NearestPowerOfTwo(Ndata);
    OneDData = DynArray_MakeRealArray(Ndata);   /*allocate the array */




    argplace = Util_ArgvStringExists("-B");
    if (argplace > 0) {   /* found binary file */
	doublearray = DynArray_MakeDoubleArray(Ndata);

	if (binfile != NULL)
	    binfile = freopen(strcpy(STR1, P_argv[argplace + 1]), "rb",
			      binfile);
	else
	    binfile = fopen(strcpy(STR1, P_argv[argplace + 1]), "rb");
	if (binfile == NULL)
	    _EscIO(FileNotFound);
	fread(doublearray, sizeof(double) * Ndata, 1, binfile);
	FORLIM = Ndata;
	for (i = 0; i < FORLIM; i++)
	    OneDData[i] = doublearray[i];
	DynArray_FreeArray((void **)(&doublearray));
    } else {
	argplace = Util_ArgvStringExists("-A");
	if (argplace > 0) {
	    if (asciifile != NULL)
		asciifile = freopen(strcpy(STR1, P_argv[argplace + 1]), "r",
				    asciifile);
	    else
		asciifile = fopen(strcpy(STR1, P_argv[argplace + 1]), "r");
	    if (asciifile == NULL)
		_EscIO(FileNotFound);
	    FORLIM = Ndata;
	    for (i = 0; i < FORLIM; i++) {
		fscanf(asciifile, "%lg", &number);
		OneDData[i] = number;
	    }
	} else
	    Util_FatalError(
	      "You must provide either a -A (ascii) or -B (binary) file");
    }

    if (binfile != NULL)
	fclose(binfile);
    if (asciifile != NULL)
	fclose(asciifile);
}


Static void nvc_GetOptions(void)
{
    /*
     *  Get the other pieces of data from the command line :
     *
     *  -d dim1 dim2
     */
    unsigned long argplace;
    Char TEMP[256];

    if (P_argc == 1) {
	IO_WritelnS(&_iob[2], 1, 6, "usage:");
	IO_WritelnS(&_iob[2], 1, 68,
	  "nvc -N npoints(power-of-2) -delay tmin tmax dt -pred prediction_time");
	IO_WritelnS(&_iob[2], 1, 45,
		    "    -correl decorrelation_time -dim dmin dmax");
	IO_WritelnS(&_iob[2], 1, 64,
	  "    -nfake number_of_fake_datasets -decimate decimation_interval");

	IO_WritelnS(&_iob[2], 1, 65,
	  "-o output_file {-B input_datafile.dbin}|{-A input_datafile.ascii}");
	IO_WritelnS(&_iob[2], 1, 44,
		    "Optional:  -phasesonly, -dumpdata, -dumperrs");

	exit(0);
    }

    argplace = Util_ArgvStringExists("-dim");
    if (argplace == 0)
	Util_FatalError("You must provide a -dim min_dim max_dim flag!");
    strcpy(TEMP, P_argv[argplace + 1]);
    Dmin = Util_ScanInteger(TEMP);
    strcpy(TEMP, P_argv[argplace + 2]);
    Dmax = Util_ScanInteger(TEMP);

    argplace = Util_ArgvStringExists("-delay");
    if (argplace == 0)
	Util_FatalError("You must provide a -delay tau1 tau2 deltatau flag!");
    strcpy(TEMP, P_argv[argplace + 1]);
    Tau1 = Util_ScanInteger(TEMP);
    strcpy(TEMP, P_argv[argplace + 2]);
    Tau2 = Util_ScanInteger(TEMP);
    strcpy(TEMP, P_argv[argplace + 3]);
    DTau = Util_ScanInteger(TEMP);


    argplace = Util_ArgvStringExists("-correl");
    if (argplace == 0)
	Util_FatalError("You must provide a -correl decorrelation_time flag!");
    strcpy(TEMP, P_argv[argplace + 1]);
    Correl = Util_ScanInteger(TEMP);

    argplace = Util_ArgvStringExists("-nfake");
    if (argplace == 0)
	Util_FatalError("You must provide a -nfake flag!");
    strcpy(TEMP, P_argv[argplace + 1]);
    Nfake = Util_ScanInteger(TEMP);

    argplace = Util_ArgvStringExists("-pred");
    if (argplace == 0)
	Util_FatalError("You must provide a -pred flag!");
    strcpy(TEMP, P_argv[argplace + 1]);
    PredTime = Util_ScanInteger(TEMP);

    argplace = Util_ArgvStringExists("-decimate");
    if (argplace == 0)
	Util_FatalError("You must provide a -decimate decorrelation_time flag!");
    strcpy(TEMP, P_argv[argplace + 1]);
    Decimate = Util_ScanInteger(TEMP);

    if (Util_ArgvStringExists("-phasesonly"))
	phasesonly = true;
    else
	phasesonly = false;

    if (Util_ArgvStringExists("-dumpdata"))
	dumpdata = true;
    else
	dumpdata = false;

    if (Util_ArgvStringExists("-dumperrs"))
	dumperrs = true;
    else
	dumperrs = false;

    argplace = Util_ArgvStringExists("-o");
    if (argplace == 0)
	strcpy(outfilename, "nvc.out");
    else
	strcpy(outfilename, P_argv[argplace + 1]);
    if (outfile != NULL)
	outfile = freopen(outfilename, "w", outfile);
    else
	outfile = fopen(outfilename, "w");
    if (outfile == NULL)
	_EscIO(FileNotFound);





}


Static unsigned long nvc_EmbedData2(float **m, float *pred, float *oned,
  unsigned long ndata, unsigned long d, unsigned long tau,
  unsigned long predtime)
{
    /*
     *  Do the time-delay embedding, and find the 'want to predict'
     *  component.
     *  returning the number of valid vectors into nvec.
     */
    unsigned long i, j, nvec;


    nvec = ndata - d * tau - predtime;

    for (i = 0; i < nvec; i++) {
	for (j = 0; j < d; j++)
	    m[i][j] = oned[i + j * tau];

	pred[i] = oned[i + (d - 1) * tau + predtime];

    }

    return nvec;

}


Static void nvc_NormalizeData(void)
{
    unsigned long i;
    double m, Size, d;
    unsigned long FORLIM;
    double TEMP;


    m = 0.0;
    FORLIM = Ndata;
    for (i = 0; i < FORLIM; i++)
	m += OneDData[i];
    m /= Ndata;

    d = 0.0;
    FORLIM = Ndata;
    for (i = 0; i < FORLIM; i++) {
	TEMP = OneDData[i] - m;
	d += TEMP * TEMP;
    }

    Size = sqrt(d / Ndata);

    FORLIM = Ndata;
    for (i = 0; i < FORLIM; i++)
	OneDData[i] = (OneDData[i] - m) / Size;

}



extern float nvc_NormalDeviate(long *iseed);

/* from numerical recipes */



extern float nvc_UniformDeviate(long *iseed);


Static void nvc_TwoNormalDeviates(long *iseed, float *f1, float *f2)
{
    /* return TWO normal deviates, with gaussian distribution
       and ***1/2*** variance, to be used in the data faking
       routine
     */
    float r, fac, v1, v2;


    do {
	v1 = 2.0 * nvc_UniformDeviate(iseed) - 1.0;
	v2 = 2.0 * nvc_UniformDeviate(iseed) - 1.0;
	r = v1 * v1 + v2 * v2;
    } while (r >= 1.0);

    fac = sqrt(-(log((double)r) / r));

    *f1 = v1 * fac;
    *f2 = v2 * fac;

}


Static void nvc_CreateGaussianizedData(unsigned long N, float *in, float *out)
{
    /*
     *  Do a 'histogram' transformation of the 'in' data set
     *  so that the 'out' data set has gaussian p.d.f. but everything
     *  else is essentially the same.
     */
    float *index_, *incopy;

    unsigned long i;


    index_ = DynArray_MakeRealArray(N);
    incopy = DynArray_MakeRealArray(N);

    for (i = 0; i < N; i++) {
	incopy[i] = in[i];
	out[i] = nvc_NormalDeviate(&iseed);
	index_[i] = (float)i;
    }  /* FOR i */

    Math_HeapSortReal(N, out);
    Math_HeapSortReal2(N, incopy, index_);
    Math_HeapSortReal2(N, index_, out);

    DynArray_FreeArray((void **)(&index_));
    DynArray_FreeArray((void **)(&incopy));

}


/* Forward FFT */

extern void nvc_RealFFT(unsigned long N, float *data);


/* Inverse FFT */

extern void nvc_RealIFFT(unsigned long N, float *data);


Static float nvc_KolmogorovSmirnoffHelper(float lambda)
{
    /* returns LOG10 of the probability that sets are same, with
       lambda = KS-stat * (N1*N2)/(N1+N2) */
    unsigned long j;
    float a2, fac, sum, term, termbf;

    fac = 2.0;
    sum = 0.0;
    termbf = 0.0;
    a2 = -2.0 * lambda * lambda;

    if (a2 < -200)   /* approximation to avoid blowup */
	return (log10(2.0) + a2 * log10(2.7182818));

    for (j = 1; j <= 100; j++) {
	term = fac * exp(a2 * (double)(j * j));
	sum += term;
	if (fabs(term) <= 0.001 * termbf || fabs(term) < 1.0e-8 * sum)
	    return log10(sum);

	fac = -fac;
	termbf = fabs(term);

    }  /* for j */
    return 0.0;

}


Static void nvc_DumpMeanStdDev(unsigned long N, float *data)
{
    unsigned long i;
    float sum, mean, dev;
    double TEMP;


    sum = 0.0;
    for (i = 0; i < N; i++)
	sum += data[i];
    mean = sum / (float)N;

    sum = 0.0;
    for (i = 0; i < N; i++) {
	TEMP = data[i] - mean;
	sum += TEMP * TEMP;
    }
    dev = sqrt(sum / (float)N);

    printf("%15.10f%15.10f\n", mean, dev);

}


Static boolean nvc_PassKolmogorovSmirnoffTest(unsigned long N, float *real,
					      float *fake)
{
    /*
     * Return TRUE if the two data sets have 1-dim p/d/f's
     * identical to 95% confidence
     */
    float *realcopy, *fakecopy;
    unsigned long i;

    unsigned long j1, j2;
    float fn1, fn2, d, d1, d2, dt, l10prob;

    /* After numerical recipes */


    realcopy = DynArray_MakeRealArray(N);
    fakecopy = DynArray_MakeRealArray(N);

    for (i = 0; i < N; i++)
	realcopy[i] = real[i];

    for (i = 0; i < N; i++)
	fakecopy[i] = fake[i];

    /* now compute KS test */
    Math_HeapSortReal(N, realcopy);
    Math_HeapSortReal(N, fakecopy);


    j1 = 0;
    j2 = 0;
    fn1 = 0.0;
    fn2 = 0.0;
    d = 0.0;

    while (j1 < N && j2 < N) {
	d1 = realcopy[j1];
	d2 = fakecopy[j2];

	if (d1 <= d2) {
	    fn1 = (float)j1 / (float)N;
	    j1++;
	}

	if (d2 <= d1) {
	    fn2 = (float)j2 / (float)N;
	    j2++;
	}

	dt = fabs(fn2 - fn1);
	if (dt > d)
	    d = dt;

    }  /* while */

    l10prob = nvc_KolmogorovSmirnoffHelper(d * sqrt((double)N / 2.0));
    printf("Log10 KS test = %10.5f\n", l10prob);

    DynArray_FreeArray((void **)(&realcopy));
    DynArray_FreeArray((void **)(&fakecopy));

    if (l10prob >= -1.310)
	return true;
    else
	return false;

}


Static float nvc_MannWhitneyStat(unsigned long N1, float *data1,
				 unsigned long N2, float *data2)
{
    /*
     *  Do a mann-whitney test on the ABSOLUTE VALUES of the elements
     *  of data1 and data2
     */
    unsigned long i, N;
    float rn1, rn2, stat;
    float *bigdata, *which;
    float sum1, sum2;

    unsigned long rankx, yindex;

    N = N1 + N2;

    bigdata = DynArray_MakeRealArray(N);
    which = DynArray_MakeRealArray(N);

    sum1 = 0.0;
    for (i = 0; i < N1; i++) {
	bigdata[i] = fabs(data1[i]);
	sum1 += bigdata[i];
	which[i] = 0.0;   /* from data set 1*/
    }

    sum2 = 0.0;
    for (i = 0; i < N2; i++) {
	bigdata[i + N1] = fabs(data2[i]);
	sum2 += bigdata[i + N1];
	which[i + N1] = 1.0;   /* from data set 2*/
    }

    /* sort the whole shebang */
    Math_HeapSortReal2(N, bigdata, which);

    rankx = 0;
    yindex = 1;
    for (i = 0; i < N; i++) {
	if (which[i] == 0.0)
	    rankx += yindex;
	else
	    yindex++;

    }

    rn1 = (float)N1;
    rn2 = (float)N2;
    stat = (float)rankx - rn1 * rn2 / 2.0;
    stat /= sqrt(rn1 * rn2);
    stat /= sqrt((rn1 + rn2 + 1.0) / 12.0);


    printf("*** in ranktest: mean err1 = % .5E err2 = % .5E\n",
	   sum1 / N1, sum2 / N2);

    DynArray_FreeArray((void **)(&bigdata));
    DynArray_FreeArray((void **)(&which));

    return stat;

}


Static void nvc_CreateFakeData(unsigned long N, float *real_data,
			       float *fake_data)
{
    unsigned long i;
    float a, b, r, fr, fi;


    /* copy real data into fake_data */
    do {
	for (i = 0; i < N; i++)
	    fake_data[i] = real_data[i];

	nvc_RealFFT(N, fake_data);
	/* compute real FFT in place.  In numerical recipes format */

	for (i = 2; i <= N - 2; i += 2) {
	    nvc_TwoNormalDeviates(&iseed, &a, &b);
	    /* complex random number = a + bi */

	    if (phasesonly) {
		r = sqrt(a * a + b * b);
		a /= r;
		b /= r;
		/* a and b are now on unit circle*/
	    }

	    fr = fake_data[i];
	    fi = fake_data[i + 1];

	    fake_data[i] = a * fr - b * fi;
	    fake_data[i + 1] = a * fi + b * fr;
	    /* complex multiplication */

	}  /* For i */

	nvc_RealIFFT(N, fake_data);

    } while (nvc_PassKolmogorovSmirnoffTest(N, real_data, fake_data) != true);

}


Static void nvc_ComputePredictorErrors(unsigned long N, unsigned long d,
				       float **data, float *pred, float *errs)
{
    unsigned long i;
    unsigned long *indexes;
    float *distances;
    tTopRec *tree;


    indexes = DynArray_MakeCardArray(10);
    distances = DynArray_MakeRealArray(10);

    tree = KdTree_CreateKdTree(data, N, d);


    for (i = 0; i < N; i++) {
	/* local constant predictor. Find nearest neighbor
	   to data^[i].
	 */
	KdTree_SearchKdTreeCorrel(tree, data[i], 2, i, Correl, indexes,
				  distances);

	errs[i] = pred[indexes[0]] - pred[i];

    }  /* FOR i */

    putchar('\n');
    KdTree_DestroyKdTree(&tree);
    DynArray_FreeArray((void **)(&indexes));
    DynArray_FreeArray((void **)(&distances));

}


Static void nvc_DumpData(Char *s, unsigned long N, float *d1, float *d2)
{
    unsigned long i;
    FILE *f;

    f = NULL;
    if (f != NULL)
	f = freopen(s, "w", f);
    else
	f = fopen(s, "w");
    if (f == NULL)
	_EscIO(FileNotFound);
    for (i = 0; i < N; i++)
	fprintf(f, "%15.10f%15.10f\n", d1[i], d2[i]);
    if (f != NULL)
	fclose(f);
    f = NULL;

    if (f != NULL)
	fclose(f);
}


Static float nvc_ComputeZstatistic(unsigned long N, unsigned long t,
				   unsigned long d, float *real_data)
{
    float *real_errs, *real_errs_decimated, *fake_errs, *fake_errs_decimated;

    float *fake_data;
    float **embedded_data;
    float *pred_data;

    unsigned long nv1, nv2;

    unsigned long i, j, nr, nf;

    float result;
    unsigned long FORLIM;


    embedded_data = DynMatrix_MakeMatrix(Ndata, d);
    pred_data = DynArray_MakeRealArray(Ndata);
    fake_data = DynArray_MakeRealArray(Ndata);

    real_errs = DynArray_MakeRealArray(Ndata);
    fake_errs = DynArray_MakeRealArray(Ndata);
    /*    real_errs_decimated := MakeRealArray( (Ndata / Decimate) + 1);
        fake_errs_decimated := MakeRealArray( ((Ndata / Decimate) +1)*Nfake);
     */
    real_errs_decimated = DynArray_MakeRealArray(Ndata);
    fake_errs_decimated = DynArray_MakeRealArray(Ndata * Nfake);

    nr = 0;
    nf = 0;

    printf("Embedding data.\n");
    nv1 = nvc_EmbedData2(embedded_data, pred_data, real_data, Ndata, d, t,
			 PredTime);
    printf("Computing real predictor errors.\n");
    nvc_ComputePredictorErrors(nv1, d, embedded_data, pred_data, real_errs);

    /* now decimate these */
    i = 0;
    while (i < nv1) {
	real_errs_decimated[nr] = real_errs[i];
	nr++;
	i += Decimate;
    }

    FORLIM = Nfake;
    for (j = 1; j <= FORLIM; j++) {
	printf("Creating fake data #%12lu\n", j);
	nvc_CreateFakeData(Ndata, real_data, fake_data);
	if (dumpdata) {
	    printf("DUMPING DATA!\n");
	    nvc_DumpData("DUMPDATA.dat", Ndata, real_data, fake_data);
	}

	printf("Embedding fake data\n");
	nv2 = nvc_EmbedData2(embedded_data, pred_data, fake_data, Ndata, d, t,
			     PredTime);
	printf("Computing fake predictor errors.\n");
	nvc_ComputePredictorErrors(nv2, d, embedded_data, pred_data,
				   fake_errs);
	if (dumperrs) {
	    printf("DUMPING ERRORS!\n");
	    nvc_DumpData("DUMPERRS.dat", Ndata, real_errs, fake_errs);
	}

	i = 0;
	while (i < nv2) {
	    fake_errs_decimated[nf] = fake_errs[i];
	    nf++;
	    i += Decimate;
	}

    }  /* for j */


    /* Now we have the two sets of errors */
    /* do mann-whitney test here */
    printf("Computing statistic\n");
    result = nvc_MannWhitneyStat(nr, real_errs_decimated, nf,
				 fake_errs_decimated);

    DynArray_FreeArray((void **)(&fake_errs_decimated));
    DynArray_FreeArray((void **)(&real_errs_decimated));
    DynArray_FreeArray((void **)(&fake_errs));
    DynArray_FreeArray((void **)(&real_errs));

    DynArray_FreeArray((void **)(&pred_data));
    DynArray_FreeArray((void **)(&fake_data));

    DynMatrix_FreeMatrix(&embedded_data, Ndata, d);

    return result;

}



Static void nvc_ProcessData(void)
{
    unsigned long d;   /* present dim */
    unsigned long t;   /* present timedelay */
    float *gd;   /* Gaussianized original data */
    float stat;
    unsigned long FORLIM, FORLIM1;


    gd = DynArray_MakeRealArray(Ndata);

    FORLIM = Tau2;
    for (t = Tau1; t <= FORLIM; t += DTau) {
	fprintf(outfile, "%10lu ", t);
	FORLIM1 = Dmax;
	for (d = Dmin; d <= FORLIM1; d++) {
	    printf("Tau = %12lu d = %12lu\n", t, d);
	    printf("gaussianizing data\n");
	    nvc_CreateGaussianizedData(Ndata, OneDData, gd);
	    stat = nvc_ComputeZstatistic(Ndata, t, d, gd);
	    printf("**** Statistic is %15.8f\n", stat);
	    fprintf(outfile, "%15.8f ", stat);
	    IO_Flush(outfile);
	}

	/* Transform to gaussian PDF */
	putc('\n', outfile);
    }  /* FOR t */

    DynArray_FreeArray((void **)(&gd));

}



main(int argc, Char *argv[])
{  /* the main program */
    PASCAL_MAIN(argc, argv);
    outfile = NULL;
    iseed = -117;

    nvc_GetOptions();
    nvc_ReadInData();
    nvc_NormalizeData();

    nvc_ProcessData();


    if (outfile != NULL)
	fclose(outfile);
    exit(EXIT_SUCCESS);
}





/* End. */
