MODULE nvc;
(*
 *
 *      Noise vs. Chaos code.
 *
 *   This code implements a test to compare a given signal
 *   to the hypothesis that the signal is colored noise.  The test
 *   is designed to detect predictability in low dimensional dynamical
 *   systems, so if the test fails, the data may be from a
 *   chaotic system.  (Though not always! The test can be failed by
 *   nonlinear noise, I think, and periodic signals plus added noise)

 *   The method is in:
 *   M. B. Kennel and S. Isabelle, ``Method to distinguish possible
 *   chaos from colored noise and determine embedding parameters'',
 *   to be published in September 15th,1992 issue of Physical Review A.
 *   Preprint also available by anonymous FTP from 'lyapunov.ucsd.edu'.

 *      Usage:
 *
 *      nvc -N Npoints(Power of 2)
 *          -delay time_delay_min time_delay_max time_delay_step
 *          -pred prediction_time
 *          -correl decorrelation_dime
 *          -decimate prediction_errors_decimation_interval
 *          -nfake number_of_fake_datasets
 *          -dim embedding_dimension_min embedding_dimension_max

	(all those parameters are integers)

 *          -o output_file_name
 *
 *          -A datafile.ascii | -B datafile.dbin
 *      (input files, either in ASCII or double precision binary FP)
 *
 *      Optional:
 *           -phasesonly         Only randomize phases to make fake data,
 *			         not amplitude.
	     -dumpdata           Dump the gaussianized 1-d real and fake
				 datasets to DUMP.dat--EVERYTIME
	     -dumperrs		 Dump the predictor errors to DUMPERRS.dat
 *)

FROM    DynArray        IMPORT
    MakeRealArray,CardArrayPtr,RealArrayPtr;

FROM    DynMatrix       IMPORT
    MakeMatrix, RealMatrixPtr;

FROM    Util            IMPORT
    ScanInteger, ArgvStringExists, FatalError, Exit;

FROM    KdTree          IMPORT
    KdTreePtr,CreateKdTree,SearchKdTree;

FROM    Math            IMPORT
    HeapSortReal;

FROM    SYSTEM          IMPORT
    LN, EXP, SQRT;

IMPORT
    IO;


VAR

    OneDData        :   RealArrayPtr; (* The 1-D data array *)

    Ndata           :   CARDINAL;       (* the number of data points *)

    iseed	    :	INTEGER;


    Tau1, Tau2, DTau:   CARDINAL;
    PredTime        :   CARDINAL;
    Dmin, Dmax      :   CARDINAL;
    Correl          :   CARDINAL;
    Decimate        :   CARDINAL;
    Nfake           :   CARDINAL;

    outfile         :   TEXT;
    outfilename     :   STRING;

    phasesonly	    :	BOOLEAN;    (* do we want to randomize phases only *)
    dumpdata	    :	BOOLEAN;    (* Dump data sets *)
    dumperrs	    :	BOOLEAN;    (* Dump errors *)

PROCEDURE NearestPowerOfTwo(N   :	CARDINAL) : CARDINAL;

VAR
    i, j, k	    :	CARDINAL;

BEGIN
    j := N;
    i := 0;
    REPEAT
	j := j / 2;
	i := i + 1;
    UNTIL (j = 1);

    FOR k := 1 TO i DO
	j := j * 2;
    END;

    RETURN(j);

END;

PROCEDURE   ReadInData;
(*
 *    From the arguments passed to the program get the input data vector
 *    Either from an ascii file, or from a binary file:
 *
 *    fnn -A filename.ascii
 *   or
 *    fnn -B filename.binary
 *
 *    -B is searched for FIRST on the command line
 *    Allocates OneDData;
 *)

VAR

    doublearray	    :	DoubleArrayPtr;
    argplace        :   CARDINAL;
    binfile         :   FILE;
    asciifile       :   TEXT;
    i               :   CARDINAL;
    number	    :	LONGREAL;
BEGIN

    argplace := ArgvStringExists("-N");
    IF (argplace = 0) THEN
	FatalError("You must provide a -N npoints flag!");
    END;

    Ndata := ScanInteger(PARAMSTR(argplace+1));

    writeln('Ndata = ',Ndata, ' NearestPowerOfTwo(Ndata) = ',
	    NearestPowerOfTwo(Ndata) );
    Ndata := NearestPowerOfTwo(Ndata);
    OneDData := MakeRealArray(Ndata); (*allocate the array *)




    argplace := ArgvStringExists("-B");
    IF (argplace > 0) THEN              (* found binary file *)
	doublearray := MakeDoubleArray(Ndata);

	RESET(binfile,PARAMSTR(argplace+1));
	BLOCKREAD(binfile,doublearray^,TSIZE(LONGREAL)*Ndata);
	FOR i := 0 TO Ndata-1 DO
	    OneDData^[i] := doublearray^[i];
	END;
	FreeArray(doublearray);
    ELSE
	argplace := ArgvStringExists("-A");
	IF (argplace > 0) THEN
	    RESET(asciifile,PARAMSTR(argplace+1));
	    FOR i := 0 TO Ndata-1 DO
		READ(asciifile,number);
		OneDData^[i] := number;
	    END;
	ELSE
	    FatalError(
	    "You must provide either a -A (ascii) or -B (binary) file");
	END;
    END;

END ReadInData;


PROCEDURE GetOptions;
(*
 *  Get the other pieces of data from the command line :
 *
 *  -d dim1 dim2
 *)
VAR
    argplace        :   CARDINAL;

BEGIN
    IF (PARAMCOUNT = 0) THEN
	IO.WritelnS(IO.StdError,"usage:");
	IO.WritelnS(IO.StdError,
	"nvc -N npoints(power-of-2) -delay tmin tmax dt -pred prediction_time");
	IO.WritelnS(IO.StdError,"    -correl decorrelation_time -dim dmin dmax");
	IO.WritelnS(IO.StdError,"    -nfake number_of_fake_datasets -decimate decimation_interval");

	IO.WritelnS(IO.StdError,
	"-o output_file {-B input_datafile.dbin}|{-A input_datafile.ascii}");
	IO.WritelnS(IO.StdError,"Optional:  -phasesonly, -dumpdata, -dumperrs");

	Exit(0);
    END;

    argplace := ArgvStringExists("-dim");
    IF (argplace = 0) THEN
	FatalError("You must provide a -dim min_dim max_dim flag!");
    END;
    Dmin := ScanInteger(PARAMSTR(argplace+1));
    Dmax := ScanInteger(PARAMSTR(argplace+2));

    argplace := ArgvStringExists("-delay");
    IF (argplace = 0) THEN
	FatalError("You must provide a -delay tau1 tau2 deltatau flag!");
    END;
    Tau1 := ScanInteger(PARAMSTR(argplace+1));
    Tau2 := ScanInteger(PARAMSTR(argplace+2));
    DTau := ScanInteger(PARAMSTR(argplace+3));


    argplace := ArgvStringExists("-correl");
    IF (argplace = 0) THEN
	FatalError("You must provide a -correl decorrelation_time flag!");
    END;
    Correl := ScanInteger(PARAMSTR(argplace+1));

    argplace := ArgvStringExists("-nfake");
    IF (argplace = 0) THEN
	FatalError("You must provide a -nfake flag!");
    END;
    Nfake := ScanInteger(PARAMSTR(argplace+1));

    argplace := ArgvStringExists("-pred");
    IF (argplace = 0) THEN
	FatalError("You must provide a -pred flag!");
    END;
    PredTime := ScanInteger(PARAMSTR(argplace+1));

    argplace := ArgvStringExists("-decimate");
    IF (argplace = 0) THEN
	FatalError("You must provide a -decimate decorrelation_time flag!");
    END;
    Decimate := ScanInteger(PARAMSTR(argplace+1));

    IF (ArgvStringExists("-phasesonly")) THEN
	phasesonly := TRUE;
    ELSE
	phasesonly := FALSE;
    END;

    IF (ArgvStringExists("-dumpdata")) THEN
	dumpdata := TRUE;
    ELSE
	dumpdata := FALSE;
    END;

    IF (ArgvStringExists("-dumperrs")) THEN
	dumperrs := TRUE;
    ELSE
	dumperrs := FALSE;
    END;

    argplace := ArgvStringExists("-o");
    IF (argplace = 0) THEN
	outfilename := "nvc.out";
    ELSE
	outfilename := PARAMSTR(argplace+1);
    END;
    REWRITE(outfile, outfilename);




END GetOptions;

PROCEDURE   EmbedData2(m:	RealMatrixPtr;
		      pred:	RealArrayPtr;
		      oned:     RealArrayPtr;
		      ndata:    CARDINAL;
		      d:        CARDINAL;
		      tau:      CARDINAL;
		      predtime:	CARDINAL ) : CARDINAL;
(*
 *  Do the time-delay embedding, and find the 'want to predict'
 *  component.
 *  returning the number of valid vectors into nvec.
 *)
VAR
    i,j     :       CARDINAL;
    nvec    :	    CARDINAL;

BEGIN

    nvec := ndata - (d * tau) - predtime;

    FOR i := 0 TO nvec-1 DO
	FOR j := 0 TO d-1 DO
	    m^[i]^[j] := oned^[i + j*tau];
	END;

	pred^[i] := oned^[i + (d-1)*tau + predtime];

    END;

    RETURN(nvec);

END EmbedData2;


PROCEDURE NormalizeData;

VAR
    i             :     CARDINAL;
    m, Size,d,tmp :     LONGREAL;

BEGIN

    m := 0.0;
    FOR i := 0 TO Ndata-1 DO
	m := m + OneDData^[i];
    END;
    m := m / Ndata;

    d := 0.0;
    FOR i := 0 TO Ndata-1 DO
	d := d + SQR(OneDData^[i]-m);
    END;

    Size := SQRT(d / Ndata);

    FOR i := 0 TO Ndata-1 DO
	OneDData^[i] := (OneDData^[i] - m) / Size;
    END;

END NormalizeData;


PROCEDURE NormalDeviate(VAR iseed:  INTEGER) : REAL;

EXTERNAL;
(* from numerical recipes *)


PROCEDURE UniformDeviate(VAR iseed: INTEGER) : REAL;

EXTERNAL;

PROCEDURE TwoNormalDeviates(VAR iseed:  INTEGER;
			    VAR f1:    REAL;
			    VAR f2:    REAL );
(* return TWO normal deviates, with gaussian distribution
   and ***1/2*** variance, to be used in the data faking
   routine
 *)
VAR
    r, fac, v1, v2	:   REAL;

BEGIN

    REPEAT
	v1 := 2.0*UniformDeviate(iseed) - 1.0;
	v2 := 2.0*UniformDeviate(iseed) - 1.0;
	r := v1*v1 + v2*v2;
    UNTIL (r < 1.0);

    fac := SQRT( -LN( LONGREAL(r) ) / r );

    f1 := v1*fac;
    f2 := v2*fac;

END TwoNormalDeviates;

PROCEDURE CreateGaussianizedData(N:     CARDINAL;
				 in:    RealArrayPtr;
				 out:   RealArrayPtr );
(*
 *  Do a 'histogram' transformation of the 'in' data set
 *  so that the 'out' data set has gaussian p.d.f. but everything
 *  else is essentially the same.
 *)

VAR
    index,incopy	    :	RealArrayPtr;

    i			    :	CARDINAL;

BEGIN

    index := MakeRealArray(N);
    incopy := MakeRealArray(N);

    FOR i := 0 TO N-1 DO
	incopy^[i] := in^[i];
	out^[i] := NormalDeviate(iseed);
	index^[i] := REAL(i);
    END; (* FOR i *)

    HeapSortReal(N, out);
    HeapSortReal2(N, incopy, index);
    HeapSortReal2(N, index, out);

    FreeArray(index);
    FreeArray(incopy);

END CreateGaussianizedData;


PROCEDURE RealFFT(N : CARDINAL;
		  data : RealArrayPtr );
(* Forward FFT *)

EXTERNAL;

PROCEDURE RealIFFT(N : CARDINAL;
		  data : RealArrayPtr );

(* Inverse FFT *)

EXTERNAL;


PROCEDURE KolmogorovSmirnoffHelper(lambda:  REAL) : REAL;
(* returns LOG10 of the probability that sets are same, with
   lambda = KS-stat * (N1*N2)/(N1+N2) *)

VAR
    j	    :	CARDINAL;
    a2	    :	REAL;
    fac, sum, term, termbf : REAL;

BEGIN
    fac := 2.0;
    sum := 0.0;
    termbf := 0.0;
    a2 := -2.0*lambda*lambda;

    IF (a2 < -200) THEN	    (* approximation to avoid blowup *)
	RETURN( LOG10(2.0) + a2 * LOG10(2.7182818) );
    END;

    FOR j :=1 TO 100 DO
	term := fac * EXP(a2*LONGREAL(j*j));
	sum := sum + term;
	IF (ABS(term) <= 0.001*termbf) OR
	   (ABS(term) < 1.0e-8*sum) THEN
	    RETURN( LOG10(sum));
	END;

	fac := -fac;
	termbf := ABS(term);

    END; (* for j *)
    RETURN(0.0);

END KolmogorovSmirnoffHelper;

PROCEDURE DumpMeanStdDev(N  :	CARDINAL;
			 data:	RealArrayPtr
			 );
VAR
    i	:   CARDINAL;
    sum, mean, dev : REAL;

BEGIN

    sum := 0.0;
    FOR i := 0 TO N-1 DO
	sum := sum + data^[i];
    END;
    mean := sum / REAL(N);

    sum := 0.0;
    FOR i := 0 TO N-1 DO
	sum := sum + SQR(data^[i]-mean);
    END;
    dev := SQRT(sum / REAL(N) );

    writeln(mean:15:10, dev:15:10);

END DumpMeanStdDev;

PROCEDURE PassKolmogorovSmirnoffTest(N : CARDINAL;
				     real:  RealArrayPtr;
				     fake:  RealArrayPtr
				     ) : BOOLEAN;
(*
 * Return TRUE if the two data sets have 1-dim p/d/f's 
 * identical to 95% confidence
 *)
VAR
    realcopy	:   RealArrayPtr;
    fakecopy	:   RealArrayPtr;
    i		:   CARDINAL;

    j1, j2	:   CARDINAL;
    fn1, fn2	:   REAL;
    d,d1,d2,dt	:   REAL;
    l10prob	:   REAL;
    sum,dev	:   REAL;
(* After numerical recipes *)

BEGIN

    realcopy := MakeRealArray(N);
    fakecopy := MakeRealArray(N);

    FOR i := 0 TO N-1 DO
	realcopy^[i] := real^[i];
    END;

    FOR i := 0 TO N-1 DO
	fakecopy^[i] := fake^[i];
    END;

    (* now compute KS test *)
    HeapSortReal(N, realcopy);
    HeapSortReal(N, fakecopy);


    j1 := 0;
    j2 := 0;
    fn1 := 0.0;
    fn2 := 0.0;
    d := 0.0;

    WHILE (j1 < N) AND (j2 < N) DO
	d1 := realcopy^[j1];
	d2 := fakecopy^[j2];

	IF (d1 <= d2) THEN
	    fn1 := REAL(j1) / REAL(N);
	    j1 := j1 + 1;
	END;

	IF (d2 <= d1) THEN
	    fn2 := REAL(j2) / REAL(N);
	    j2 := j2 + 1;
	END;

	dt := ABS(fn2-fn1);
	IF (dt > d) THEN
	    d := dt;
	END;

    END; (* while *)

    l10prob := KolmogorovSmirnoffHelper( d * SQRT( LONGREAL(N)/2.0) ) ;
    writeln('Log10 KS test = ', l10prob:10:5);

    FreeArray(realcopy);
    FreeArray(fakecopy);

    IF (l10prob >= -1.310) THEN
	RETURN(TRUE);
    ELSE
	RETURN(FALSE);
    END;

END;

PROCEDURE MannWhitneyStat(N1	:   CARDINAL;
			  data1	:   RealArrayPtr;
			  N2	:   CARDINAL;
			  data2	:   RealArrayPtr
			  ) : REAL;
(*
 *  Do a mann-whitney test on the ABSOLUTE VALUES of the elements
 *  of data1 and data2
 *)

VAR
    i,N		:	CARDINAL;
    rn1, rn2	:	REAL;
    stat	:	REAL;
    bigdata	:	RealArrayPtr;
    which	:	RealArrayPtr;
    sum1,sum2	:	REAL;

    rankx, yindex:	CARDINAL;

BEGIN
    N := N1 + N2;

    bigdata := MakeRealArray(N);
    which := MakeRealArray(N);

    sum1 := 0.0;
    FOR i := 0 TO N1-1 DO
	bigdata^[i] := ABS(data1^[i]);
	sum1 := sum1 + bigdata^[i];
	which^[i] := 0.0;	(* from data set 1*)
    END;

    sum2 := 0.0;
    FOR i := 0 TO N2-1 DO
	bigdata^[i+N1] := ABS(data2^[i]);
	sum2 := sum2 + bigdata^[i+N1];
	which^[i+N1] := 1.0;	(* from data set 2*)
    END;

    (* sort the whole shebang *)
    HeapSortReal2(N, bigdata, which);

    rankx := 0;
    yindex := 1;
    FOR i := 0 TO N-1 DO
	IF (which^[i] = 0.0) THEN
		rankx := rankx + yindex;
	ELSE
		yindex := yindex+1;
	END;

    END; 

    rn1 := REAL(N1);
    rn2 := REAL(N2);
    stat :=   (REAL(rankx) - (rn1*rn2)/2.0);
    stat := stat/ SQRT(rn1*rn2);
    stat := stat / SQRT((rn1 + rn2 + 1.0)/12.0);
    

    writeln('*** in ranktest: mean err1 = ', sum1/N1,
	    ' err2 = ', sum2/N2);

    FreeArray(bigdata);
    FreeArray(which);

    RETURN(stat);

END MannWhitneyStat;

PROCEDURE CreateFakeData(N	    :	CARDINAL;
			 real_data  :	RealArrayPtr;
			 fake_data  :	RealArrayPtr );

VAR i	    :	CARDINAL;
    a, b,r  :	REAL;
    fr, fi  :	REAL;

BEGIN

    (* copy real data into fake_data *)
    REPEAT
      FOR i := 0 TO N-1 DO
	  fake_data^[i] := real_data^[i];
      END;

      RealFFT(N, fake_data);
    (* compute real FFT in place.  In numerical recipes format *)

      FOR i := 2 TO N-2 BY 2 DO
	  TwoNormalDeviates(iseed, a, b);
	(* complex random number = a + bi *)

	  IF (phasesonly) THEN
	    r := SQRT(a*a+b*b);
	    a := a / r;
	    b := b/r;
	    (* a and b are now on unit circle*)
	  END;

	  fr := fake_data^[i];
	  fi := fake_data^[i+1];

	  fake_data^[i] := a*fr - b*fi;
	  fake_data^[i+1] := a*fi + b*fr;
	(* complex multiplication *)
    
      END; (* For i *)

      RealIFFT(N, fake_data);

    UNTIL (PassKolmogorovSmirnoffTest(N, real_data, fake_data) = TRUE);
    
END CreateFakeData;


PROCEDURE ComputePredictorErrors(N:	CARDINAL;
				 d:	CARDINAL;
				 data:	RealMatrixPtr;
				 pred:	RealArrayPtr;
				 errs:	RealArrayPtr
				 );

VAR i		:	    CARDINAL;
    indexes	:	    CardArrayPtr;
    distances	:	    RealArrayPtr;
    tree	:	    KdTreePtr;

BEGIN

    indexes := MakeCardArray(10);
    distances := MakeRealArray(10);

    tree := CreateKdTree(data, N, d);


    FOR i := 0 TO N-1 DO
	(* local constant predictor. Find nearest neighbor
	   to data^[i].
	 *)
	SearchKdTreeCorrel(tree, data^[i], 2,
			   i, Correl,
			   indexes, distances);

	errs^[i] := pred^[indexes^[0]] - pred^[i];

    END; (* FOR i *)

    writeln;
    DestroyKdTree(tree);
    FreeArray(indexes);
    FreeArray(distances);

END ComputePredictorErrors;

PROCEDURE DumpData(s:	STRING;
		   N:	CARDINAL;
		   d1:	RealArrayPtr;
		   d2:	RealArrayPtr );
VAR
    i	:   CARDINAL;
    f	:   TEXT;
BEGIN

    rewrite(f, s);
    FOR i := 0 TO N-1 DO
	writeln(f,d1^[i]:15:10, d2^[i]:15:10);
    END;
    close(f);

END DumpData;

PROCEDURE ComputeZstatistic(N:	    CARDINAL;
			    t:	    CARDINAL;
			    d:	    CARDINAL;
			    real_data:   RealArrayPtr ) : REAL;

VAR
	real_errs           	:	RealArrayPtr;
	real_errs_decimated 	:	RealArrayPtr;
	fake_errs		:	RealArrayPtr;
	fake_errs_decimated	:	RealArrayPtr;

	fake_data             	:	RealArrayPtr;		
	embedded_data		:	RealMatrixPtr;
	pred_data		:	RealArrayPtr;

	nv1,nv2			:	CARDINAL;

	i,j,k			:	CARDINAL;
	nr, nf			:	CARDINAL;

	result			:	REAL;

BEGIN

    embedded_data := MakeMatrix(Ndata, d);
    pred_data := MakeRealArray(Ndata);
    fake_data := MakeRealArray(Ndata);

    real_errs := MakeRealArray(Ndata);
    fake_errs := MakeRealArray(Ndata);
(*    real_errs_decimated := MakeRealArray( (Ndata / Decimate) + 1);
    fake_errs_decimated := MakeRealArray( ((Ndata / Decimate) +1)*Nfake);
 *)
    real_errs_decimated := MakeRealArray( Ndata);
    fake_errs_decimated := MakeRealArray( Ndata*Nfake);

    nr := 0;
    nf := 0;

    writeln('Embedding data.');
    nv1 := EmbedData2(embedded_data, pred_data, real_data,
		      Ndata, d, t, PredTime);
    writeln('Computing real predictor errors.');
    ComputePredictorErrors(nv1, d, embedded_data, pred_data, real_errs);

    (* now decimate these *)
    i := 0;
    WHILE (i < nv1) DO
	real_errs_decimated^[nr] := real_errs^[i];
	nr := nr+1;
	i := i + Decimate;
    END;

    FOR j := 0 TO Nfake-1 DO

	writeln('Creating fake data #',j+1);
	CreateFakeData(Ndata, real_data, fake_data);
	IF (dumpdata) THEN
	    writeln('DUMPING DATA!');
	    DumpData("DUMPDATA.dat", Ndata, real_data, fake_data);	
	END;

	writeln('Embedding fake data');
	nv2 := EmbedData2(embedded_data, pred_data, fake_data, Ndata, d, t, PredTime);
	writeln('Computing fake predictor errors.');
	ComputePredictorErrors(nv2, d, embedded_data, pred_data, fake_errs);
	IF (dumperrs) THEN
	    writeln('DUMPING ERRORS!');
	    DumpData("DUMPERRS.dat", Ndata, real_errs, fake_errs);	
	END;
				
	i := 0;
	WHILE (i < nv2) DO
	    fake_errs_decimated^[nf] := fake_errs^[i];
	    nf := nf+1;
	    i := i + Decimate;
	END;

    END; (* for j *)

    (* Now we have the two sets of errors *)
    (* do mann-whitney test here *)
    writeln('Computing statistic');
    result := MannWhitneyStat(nr, real_errs_decimated,
			      nf, fake_errs_decimated);

    FreeArray(fake_errs_decimated);
    FreeArray(real_errs_decimated);
    FreeArray(fake_errs);
    FreeArray(real_errs);

    FreeArray(pred_data);
    FreeArray(fake_data);

    FreeMatrix(embedded_data, Ndata, d);

    RETURN(result);

END ComputeZstatistic;



PROCEDURE ProcessData;


VAR
    d:	    CARDINAL;                       (* present dim *)
    t:	    CARDINAL;                       (* present timedelay *)
    gd:	    RealArrayPtr;                   (* Gaussianized original data *)
    stat:   REAL;

BEGIN

    gd := MakeRealArray(Ndata);

    FOR t:=Tau1 TO Tau2 BY DTau DO
	write(outfile, t:10:5,' ');
	FOR d := Dmin TO Dmax DO

					(* Transform to gaussian PDF *)
	    writeln('Tau = ', t, ' d = ', d);
	    writeln('gaussianizing data');
	    CreateGaussianizedData(Ndata, OneDData, gd);
	    stat := ComputeZstatistic(Ndata, t, d, gd);
	    writeln('**** Statistic is ', stat:15:8);
	    write(outfile, stat:15:8, ' ');
	    IO.Flush(outfile);
	END;
	writeln(outfile);
    END;                                (* FOR t *)

    FreeArray(gd);

END ProcessData;



BEGIN                                   (* the main program *)

    iseed := -117;

    GetOptions;
    ReadInData;
    NormalizeData;

    ProcessData;


END nvc.


